{-# OPTIONS -#include "PrelPosix_hsc.h" #-}
{-# LINE 1 "PrelPosix.hsc" #-}
{-# OPTIONS -fno-implicit-prelude #-}
{-# LINE 2 "PrelPosix.hsc" #-}

-- ---------------------------------------------------------------------------
--
-- POSIX support layer for the standard libraries
--
-- Non-posix compliant in order to support the following features:
--	* S_ISSOCK (no sockets in POSIX)

module PrelPosix where

-- See above comment for non-Posixness reasons.
-- #include "PosixSource.h"


{-# LINE 16 "PrelPosix.hsc" #-}

import PrelBase
import PrelNum
import PrelReal
import PrelMaybe
import PrelCString
import PrelPtr
import PrelWord
import PrelInt
import PrelCTypesISO
import PrelCTypes
import PrelCError
import PrelStorable
import PrelMarshalAlloc
import PrelMarshalUtils
import PrelBits
import PrelIOBase
import Monad


-- ---------------------------------------------------------------------------
-- Types

data CDir    = CDir
type CSigset = ()

type CDev    = Word64
{-# LINE 43 "PrelPosix.hsc" #-}
type CIno    = Word32
{-# LINE 44 "PrelPosix.hsc" #-}
type CMode   = Word32
{-# LINE 45 "PrelPosix.hsc" #-}
type COff    = Int32
{-# LINE 46 "PrelPosix.hsc" #-}
type CPid    = Int32
{-# LINE 47 "PrelPosix.hsc" #-}


{-# LINE 51 "PrelPosix.hsc" #-}
type CGid    = Word32
{-# LINE 52 "PrelPosix.hsc" #-}
type CNlink  = Word32
{-# LINE 53 "PrelPosix.hsc" #-}
type CSsize  = Int32
{-# LINE 54 "PrelPosix.hsc" #-}
type CUid    = Word32
{-# LINE 55 "PrelPosix.hsc" #-}
type CCc     = Word8
{-# LINE 56 "PrelPosix.hsc" #-}
type CSpeed  = Word32
{-# LINE 57 "PrelPosix.hsc" #-}
type CTcflag = Word32
{-# LINE 58 "PrelPosix.hsc" #-}

{-# LINE 59 "PrelPosix.hsc" #-}

-- ---------------------------------------------------------------------------
-- stat()-related stuff

type CStat = ()

fdFileSize :: Int -> IO Integer
fdFileSize fd = 
  allocaBytes (88) $ \ p_stat -> do
{-# LINE 68 "PrelPosix.hsc" #-}
    throwErrnoIfMinus1Retry "fileSize" $
	c_fstat (fromIntegral fd) p_stat
    c_mode <- ((\hsc_ptr -> peekByteOff hsc_ptr 16)) p_stat :: IO CMode 
{-# LINE 71 "PrelPosix.hsc" #-}
    if not (s_isreg c_mode)
	then return (-1)
	else do
    c_size <- ((\hsc_ptr -> peekByteOff hsc_ptr 44)) p_stat :: IO COff
{-# LINE 75 "PrelPosix.hsc" #-}
    return (fromIntegral c_size)

data FDType  = Directory | Stream | RegularFile
	       deriving (Eq)

-- NOTE: On Win32 platforms, this will only work with file descriptors
-- referring to file handles. i.e., it'll fail for socket FDs.
fdType :: Int -> IO FDType
fdType fd = 
  allocaBytes (88) $ \ p_stat -> do
{-# LINE 85 "PrelPosix.hsc" #-}
    throwErrnoIfMinus1Retry "fdType" $
	c_fstat (fromIntegral fd) p_stat
    c_mode <- ((\hsc_ptr -> peekByteOff hsc_ptr 16)) p_stat :: IO CMode
{-# LINE 88 "PrelPosix.hsc" #-}
    case () of
      _ |  s_isdir  c_mode  -> return Directory
        |  s_isfifo c_mode  -> return Stream
	|  s_issock c_mode  -> return Stream
	|  s_ischr  c_mode  -> return Stream
	|  s_isreg  c_mode  -> return RegularFile
	|  s_isblk  c_mode  -> return RegularFile
	| otherwise	    -> ioException ioe_unknownfiletype
    -- we consider character devices to be streams (eg. ttys),
    -- whereas block devices are more like regular files because they
    -- are seekable.

ioe_unknownfiletype = IOError Nothing UnsupportedOperation "fdType"
			"unknown file type" Nothing

foreign import "s_isreg_PrelPosix_wrap" unsafe s_isreg :: CMode -> Bool

{-# LINE 105 "PrelPosix.hsc" #-}

foreign import "s_isdir_PrelPosix_wrap" unsafe s_isdir :: CMode -> Bool

{-# LINE 108 "PrelPosix.hsc" #-}

foreign import "s_isfifo_PrelPosix_wrap" unsafe s_isfifo :: CMode -> Bool

{-# LINE 111 "PrelPosix.hsc" #-}

foreign import "s_ischr_PrelPosix_wrap" unsafe s_ischr :: CMode -> Bool

{-# LINE 114 "PrelPosix.hsc" #-}

foreign import "s_isblk_PrelPosix_wrap" unsafe s_isblk :: CMode -> Bool

{-# LINE 117 "PrelPosix.hsc" #-}


{-# LINE 119 "PrelPosix.hsc" #-}
foreign import "s_issock_PrelPosix_wrap" unsafe s_issock :: CMode -> Bool

{-# LINE 121 "PrelPosix.hsc" #-}

{-# LINE 125 "PrelPosix.hsc" #-}

-- It isn't clear whether ftruncate is POSIX or not (I've read several
-- manpages and they seem to conflict), so we truncate using open/2.
fileTruncate :: FilePath -> IO ()
fileTruncate file = do
  let flags = o_WRONLY .|. o_TRUNC
  withCString file $ \file_cstr -> do
    fd <- fromIntegral `liftM`
	    throwErrnoIfMinus1Retry "fileTruncate"
 	        (c_open file_cstr (fromIntegral flags) 0o666)
    c_close fd
  return ()

-- ---------------------------------------------------------------------------
-- Terminal-related stuff

fdIsTTY :: Int -> IO Bool
fdIsTTY fd = c_isatty (fromIntegral fd) >>= return.toBool


{-# LINE 145 "PrelPosix.hsc" #-}

type Termios = ()

setEcho :: Int -> Bool -> IO ()
setEcho fd on = do
  allocaBytes (60)  $ \p_tios -> do
{-# LINE 151 "PrelPosix.hsc" #-}
    throwErrnoIfMinus1Retry "setEcho"
	(c_tcgetattr (fromIntegral fd) p_tios)
    c_lflag <- ((\hsc_ptr -> peekByteOff hsc_ptr 12)) p_tios :: IO CTcflag
{-# LINE 154 "PrelPosix.hsc" #-}
    let new_c_lflag | on        = c_lflag .|. (8)
{-# LINE 155 "PrelPosix.hsc" #-}
	            | otherwise = c_lflag .&. complement (8)
{-# LINE 156 "PrelPosix.hsc" #-}
    ((\hsc_ptr -> pokeByteOff hsc_ptr 12)) p_tios (new_c_lflag :: CTcflag)
{-# LINE 157 "PrelPosix.hsc" #-}
    tcSetAttr fd (0) p_tios
{-# LINE 158 "PrelPosix.hsc" #-}

getEcho :: Int -> IO Bool
getEcho fd = do
  allocaBytes (60)  $ \p_tios -> do
{-# LINE 162 "PrelPosix.hsc" #-}
    throwErrnoIfMinus1Retry "setEcho"
	(c_tcgetattr (fromIntegral fd) p_tios)
    c_lflag <- ((\hsc_ptr -> peekByteOff hsc_ptr 12)) p_tios :: IO CTcflag
{-# LINE 165 "PrelPosix.hsc" #-}
    return ((c_lflag .&. (8)) /= 0)
{-# LINE 166 "PrelPosix.hsc" #-}

setCooked :: Int -> Bool -> IO ()
setCooked fd cooked = 
  allocaBytes (60)  $ \p_tios -> do
{-# LINE 170 "PrelPosix.hsc" #-}
    throwErrnoIfMinus1Retry "setCooked"
	(c_tcgetattr (fromIntegral fd) p_tios)

    -- turn on/off ICANON
    c_lflag <- ((\hsc_ptr -> peekByteOff hsc_ptr 12)) p_tios :: IO CTcflag
{-# LINE 175 "PrelPosix.hsc" #-}
    let new_c_lflag | cooked    = c_lflag .|. (2)
{-# LINE 176 "PrelPosix.hsc" #-}
	            | otherwise = c_lflag .&. complement (2)
{-# LINE 177 "PrelPosix.hsc" #-}
    ((\hsc_ptr -> pokeByteOff hsc_ptr 12)) p_tios (new_c_lflag :: CTcflag)
{-# LINE 178 "PrelPosix.hsc" #-}

    -- set VMIN & VTIME to 1/0 respectively
    when cooked $ do
	    let c_cc  = ((\hsc_ptr -> hsc_ptr `plusPtr` 17)) p_tios
{-# LINE 182 "PrelPosix.hsc" #-}
		vmin  = c_cc `plusPtr` (6)  :: Ptr Word8
{-# LINE 183 "PrelPosix.hsc" #-}
		vtime = c_cc `plusPtr` (5) :: Ptr Word8
{-# LINE 184 "PrelPosix.hsc" #-}
	    poke vmin  1
	    poke vtime 0

    tcSetAttr fd (0) p_tios
{-# LINE 188 "PrelPosix.hsc" #-}

-- tcsetattr() when invoked by a background process causes the process
-- to be sent SIGTTOU regardless of whether the process has TOSTOP set
-- in its terminal flags (try it...).  This function provides a
-- wrapper which temporarily blocks SIGTTOU around the call, making it
-- transparent.

tcSetAttr :: FD -> CInt -> Ptr Termios -> IO ()
tcSetAttr fd options p_tios = do
  allocaBytes (128) $ \ p_sigset -> do
{-# LINE 198 "PrelPosix.hsc" #-}
  allocaBytes (128) $ \ p_old_sigset -> do
{-# LINE 199 "PrelPosix.hsc" #-}
     c_sigemptyset p_sigset
     c_sigaddset   p_sigset (22)
{-# LINE 201 "PrelPosix.hsc" #-}
     c_sigprocmask (0) p_sigset p_old_sigset
{-# LINE 202 "PrelPosix.hsc" #-}
     throwErrnoIfMinus1Retry_ "tcSetAttr" $
	 c_tcsetattr (fromIntegral fd) options p_tios
     c_sigprocmask (2) p_old_sigset nullPtr
{-# LINE 205 "PrelPosix.hsc" #-}


{-# LINE 219 "PrelPosix.hsc" #-}

-- ---------------------------------------------------------------------------
-- Turning on non-blocking for a file descriptor


{-# LINE 224 "PrelPosix.hsc" #-}

setNonBlockingFD fd = do
  flags <- throwErrnoIfMinus1Retry "setNonBlockingFD"
		 (fcntl_read (fromIntegral fd) (3))
{-# LINE 228 "PrelPosix.hsc" #-}
  -- An error when setting O_NONBLOCK isn't fatal: on some systems 
  -- there are certain file handles on which this will fail (eg. /dev/null
  -- on FreeBSD) so we throw away the return code from fcntl_write.
  fcntl_write (fromIntegral fd) 
	(4) (flags .|. 2048)
{-# LINE 233 "PrelPosix.hsc" #-}

{-# LINE 239 "PrelPosix.hsc" #-}

-- -----------------------------------------------------------------------------
-- foreign imports

foreign import "stat" unsafe
   c_stat :: CString -> Ptr CStat -> IO CInt

foreign import "fstat" unsafe
   c_fstat :: CInt -> Ptr CStat -> IO CInt


{-# LINE 250 "PrelPosix.hsc" #-}
foreign import "lstat" unsafe
   c_lstat :: CString -> Ptr CStat -> IO CInt

{-# LINE 253 "PrelPosix.hsc" #-}

foreign import "open" unsafe
   c_open :: CString -> CInt -> CMode -> IO CInt

-- POSIX flags only:
o_RDONLY    = (0)	   :: CInt
{-# LINE 259 "PrelPosix.hsc" #-}
o_WRONLY    = (1)	   :: CInt
{-# LINE 260 "PrelPosix.hsc" #-}
o_RDWR      = (2)	   :: CInt
{-# LINE 261 "PrelPosix.hsc" #-}
o_APPEND    = (1024)	   :: CInt
{-# LINE 262 "PrelPosix.hsc" #-}
o_CREAT     = (64)	   :: CInt
{-# LINE 263 "PrelPosix.hsc" #-}
o_EXCL	    = (128)	   :: CInt
{-# LINE 264 "PrelPosix.hsc" #-}
o_TRUNC     = (512)	   :: CInt
{-# LINE 265 "PrelPosix.hsc" #-}


{-# LINE 270 "PrelPosix.hsc" #-}
o_NOCTTY    = (256)	   :: CInt
{-# LINE 271 "PrelPosix.hsc" #-}
o_NONBLOCK  = (2048)  :: CInt
{-# LINE 272 "PrelPosix.hsc" #-}

{-# LINE 273 "PrelPosix.hsc" #-}


{-# LINE 277 "PrelPosix.hsc" #-}

foreign import "isatty" unsafe
   c_isatty :: CInt -> IO CInt

foreign import "close" unsafe
   c_close :: CInt -> IO CInt


{-# LINE 293 "PrelPosix.hsc" #-}

foreign import "lseek" unsafe
   c_lseek :: CInt -> COff -> CInt -> IO COff

foreign import "write" unsafe 
   c_write :: CInt -> Ptr CChar -> CSize -> IO CSsize

foreign import "read" unsafe 
   c_read :: CInt -> Ptr CChar -> CSize -> IO CSsize


{-# LINE 304 "PrelPosix.hsc" #-}
foreign import "fcntl" unsafe
   fcntl_read  :: CInt -> CInt -> IO CInt

foreign import "fcntl" unsafe
   fcntl_write :: CInt -> CInt -> CInt -> IO CInt

foreign import "fork" unsafe
   fork :: IO CPid 

foreign import "sigemptyset_PrelPosix_wrap" unsafe
   c_sigemptyset :: Ptr CSigset -> IO ()

{-# LINE 316 "PrelPosix.hsc" #-}

foreign import "sigaddset" unsafe
   c_sigaddset :: Ptr CSigset -> CInt -> IO ()

foreign import "sigprocmask" unsafe
   c_sigprocmask :: CInt -> Ptr CSigset -> Ptr CSigset -> IO ()

foreign import "tcgetattr" unsafe
   c_tcgetattr :: CInt -> Ptr Termios -> IO CInt

foreign import "tcsetattr" unsafe
   c_tcsetattr :: CInt -> CInt -> Ptr Termios -> IO CInt

foreign import "unlink" unsafe 
   c_unlink :: CString -> IO CInt

foreign import "waitpid" unsafe
   c_waitpid :: CPid -> Ptr CInt -> CInt -> IO CPid

{-# LINE 335 "PrelPosix.hsc" #-}
